/**
 * @author DOGO
 * @module WebViewer3D
 * @copyright   2018 Esko Stonecube Ltd. All rights reserved.
 */

define([
    "backbone",
    "./LogMessage"
], function (
    bb,
    LogMessage
)
{
    "use strict";

    const uuid = function()
    {
        let deltaT = new Date().getTime();
        return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function(c)
        {
            const r = (deltaT + Math.random()*16)%16 | 0;
            deltaT = Math.floor(deltaT/16);
            return (c ==='x' ? r :(r&0x3|0x8)).toString(16);
        });
    };

    const GetTextureTypeKey = function(__textureType)
    {
        switch (__textureType)
        {
            case "diffuse": return "diffuse_tex";
            case "specular": return "specular_tex";
            case "internal_specular": return "int_specular_tex";
            case "normal_map": return "normal_map_tex";
            case "diffraction_grating_filter": return "diff_grating_filter_tex";
            case "diffraction_grating": return "diff_grating_tex";
            default:
                return null;
        }
    };

    const MaterialList = bb.Model.extend({

        /**
         * @alias module:WebViewer3D
         * @class MaterialList
         * Create a MaterialList from any material.json Object.
         * The assets present in this pool will either have their AssetType flag "Material" or "Texture"
         *
         * @param {Object} __attributes object containing three properties.
         * "__JSONAssetPool": An object containing materials assets (and corresponding textures assets). This has the same structure as the material.json file (not a MaterialList).
         * "frontMaterialName" if provided, replaces the "front" material key name with the provided name. This should come from the _eg_metadata.json.
         * "backMaterialName" if provided, replaces the "back" material key name with the provided name. This should come from the _eg_metadata.json.
         *
         * @constructor {MaterialList} an object containing the materials (and corresponding textures) that are used by the viewer. __JSONAssetPool attribute
         *                              has the same structure as any material.json file, except that this object has the type MaterialList augmented with
         *                              utility functions such as SetTextureURL.
         *
         *         This object can be used by functions: LoadScene, LoadSceneFromObject, and UpdateMaterials in WebViewer3DAPI
         *
         */
        initialize: function (__attributes)
        {
            if (__attributes.frontMaterialName)
                if (__attributes._JSONAssetPool.hasOwnProperty(__attributes.frontMaterialName))
                    LogMessage.Error("Asset " + __attributes.front + " is already present in the asset pool. \"front\" cannot be renamed to that name.");
                else if (__attributes._JSONAssetPool.hasOwnProperty("front"))
                {
                    __attributes._JSONAssetPool[__attributes.frontMaterialName] = __attributes._JSONAssetPool.front;
                    delete __attributes._JSONAssetPool.front;
                }

            if (__attributes.backMaterialName)
                if (__attributes._JSONAssetPool.hasOwnProperty(__attributes.backMaterialName))
                    LogMessage.Error("Asset " + __attributes.back + " is already present in the asset pool. \"back\" cannot be renamed to that name.");
                else if (__attributes._JSONAssetPool.hasOwnProperty("back"))
                {
                    __attributes._JSONAssetPool[__attributes.backMaterialName] = __attributes._JSONAssetPool.back;
                    delete __attributes._JSONAssetPool.back;
                }
        },

        /**
         * Returns an array all materials names available in this list
         * @returns {Array} array containing the different material names
         */
        GetMaterialNames: function ()
        {
            const material_names = [];
            for (let i = 0, keys = Object.keys(this.get("_JSONAssetPool")), l = keys.length; i < l; ++i)
            {
                const materialName = keys[i];
                if (this.get("_JSONAssetPool")[materialName].AssetType === "Material")
                {
                    material_names.push(materialName);
                }
            }
            return material_names;
        },

        /**
         * Adds or Replaces blank default materials to this material list. If there is already a material with one of the
         * names provided in __materialNames, it is overwritten.
         *
         * Note that in that case the texture assets that could be referred to by the overwritten material are not removed,
         * so that other materials that could refer to it stay consistent.
         *
         * This function also makes sure to clean-up the structure if there are "dangling" textures
         *
         * @param {Array} __materialNames an array of strings refering to the materials to be set to default
         */
        SetMaterialsToDefault: function (__materialNames)
        {
            for (let i = 0, keys = Object.keys(__materialNames), l = keys.length; i < l; ++i)
            {
                const materialName = __materialNames[keys[i]];
                this.get("_JSONAssetPool")[materialName] =
                    {
                        AssetType: "Material",
                        LinkIdentifer: "",//unused
                        Properties:
                            {
                                reflectance_model: "Visualizer",//unused
                                shader: null,
                                double_sided: true,
                                front:
                                    {
                                        diffuse: [0.8, 0.8, 0.8],
                                        specular: [0.7, 0.7, 0.7],
                                        ambient: [1, 1, 1],
                                        emission: [0, 0, 0],
                                        shininess: 0.6,

                                        grating_spacing_scale: 1.6,
                                        default_diff_haze: 0.8,
                                        default_diff_spacing: 0.4,
                                        double_grating: false
                                    },
                                opacity: 1,
                                absorbency: 0,//unused
                                metallic: 0//unused
                            }
                    }
            }
            //Cleanup the JSONAssetpool to remove dangling textures
            const cleanedMaterials = this.GetMaterials(this.GetMaterialNames());
            this.set("_JSONAssetPool", cleanedMaterials.get("_JSONAssetPool"));
        },

        /**
         * WARNING: USE AT YOUR OWN RISK
         * This is a convenience function that tries to determine the type of the texture given its name
         * this texture type can then be used in function SetTextureURL.
         * Note that this will work with texture names that come out of PMATE only or printable parts textures
         * that were converted using WV3DSConversion component.
         * Therefore, this does not work reliably on texture names that are generated by any other process
         * or textures that belong to non printable parts for which the user is supposed to know what type these are.
         * The fallback type will be "diffuse".
         * @param {String} __textureURL
         * @returns {String} containing the type "diffuse", "specular", "normal_map", "internal_specular", "diffraction_grating", or "diffraction_grating_filter"
         */
        GetTextureType: function (__textureURL)
        {
            if (typeof __textureURL === "string")
            {
                const filename = __textureURL.substring(__textureURL.lastIndexOf('/') + 1);//extract file name
                filename.split('.').slice(0, -1).join('.');//remove extension
                if (filename.endsWith("_calcRgba")) return "diffuse";
                if (filename.endsWith("_calcSpecRgbH")) return "specular";
                if (filename.endsWith("_calcNormalMap")) return "normal_map";
                if (filename.endsWith("_calcSpecInternalRgbH")) return "internal_specular";
                if (filename.endsWith("_calcDiffGrating")) return "diffraction_grating";
                if (filename.endsWith("_calcDiffGratingfilter")) return "diffraction_grating_filter";
            }
            return "diffuse";
        },

        /**
         * Returns an array of all possible texture types available in the given material
         * @param {String} __materialName
         * @returns {Array}
         */
        GetTextureTypes: function (__materialName)
        {
            const possible_types = ["diffuse", "specular", "normal_map", "internal_specular", "diffraction_grating", "diffraction_grating_filter"];
            const result_types = [];
            const material = this.get("_JSONAssetPool")[__materialName];

            if (!material)
            {
                LogMessage.Error("Material " + __materialName + " not found in list");
                return result_types;
            }
            if (material.AssetType !== "Material")
            {
                LogMessage.Error("Asset " + __materialName + " is not a material");
                return result_types;
            }

            for (let i = 0; i < possible_types.length; ++i)
            {
                const type = possible_types[i];
                if (GetTextureTypeKey(type) && material.Properties.front.hasOwnProperty(GetTextureTypeKey(type)))
                {
                    result_types.push(type);
                }
            }
            return result_types;
        },

        /**
         * Retrieve the texture URL of the texture of type __textureType of material __materialName
         * @param __materialName
         * @param __textureType
         * @returns {String} URL of the queried texture (or relative path)
         */
        GetTextureURL: function (__materialName, __textureType)
        {
            const material = this.get("_JSONAssetPool")[__materialName];
            if (!material)
            {
                LogMessage.Error("Material " + __materialName + " not found in list");
                return null;
            }

            const textureTypeKey = GetTextureTypeKey(__textureType);
            if (!textureTypeKey)
            {
                LogMessage.Error("Programmer: choose one of these texture types:'diffuse', 'specular', 'normal_map', 'internal_specular', 'diffraction_grating', or 'diffraction_grating_filter'");
                return null;
            }

            const textureKey = material.Properties.front[textureTypeKey];
            if (!textureKey)
            {
                LogMessage.Error("Texture of type " + __textureType + " not found for material " + __materialName + ".");
                return null;
            }
            else
            {
                const textureAsset = this.get("_JSONAssetPool")[textureKey];
                return textureAsset.Properties.filepath;
            }
        },

        /**
         * Updates the properties of the textures of a given material(__materialName)
         * The texture type type must be either "diffuse", "specular", "normal_map", "internal_specular", "diffraction_grating", or "diffraction_grating_filter"
         * If the type is omitted, then the viewer tries to determine the type of texture using its filename (from URL, this is in some cases unreliable, please
         * see GetTextureType above for details about it).
         *
         * If the texture of the given type does not exist for this material, the texture asset is then created and added to this material list.
         * User can then use the main WebViewerAPI::UpdateMaterials function to reload that material using the new texture.
         *
         * Note that this call does not remove other texture types that this material uses, so if the user wants to create a new vanilla material and assign this texture
         * to it, user should call SetMaterialsToDefault([__materialName]) first (please also refer to that function's documentation if that is your intention)
         *
         * @param __materialName
         * @param {Array} __textures array of objects, each containing the following keys:
         *                  "url": mandatory string
         *                  "type": recommended string (see documentation above)
         *                  "transparent": optional boolean indicating if the alpha channel of the texture contains at least one value < 255
         *                  "remapped_alpha_channel": optional boolean indicating if the alpha channel is mapped from [127-255] instead of [0-255] (fix for IE)
         *                  "swrap": optional string indicating wrapping mode for first texture coordinate. Can be "GL_REPEAT" (default), "GL_CLAMP", "GL_CLAMP_TO_BORDER", or "GL_CLAMP_TO_EDGE"
         *                  "twrap": optional string indicating wrapping mode for second texture coordinate. Can be "GL_REPEAT" (default), "GL_CLAMP", "GL_CLAMP_TO_BORDER", or "GL_CLAMP_TO_EDGE"
         *                  "original_width": optional original texture width before padding ( if it has been padded to a power of two for example)
         *                  "original_height": optional original texture height before padding ( if it has been padded to a power of two for example)
         */
        SetTextures: function (__materialName, __textures)
        {
            let material = this.get("_JSONAssetPool")[__materialName];
            if (!material)
            {
                this.SetMaterialsToDefault([__materialName]);
                material = this.get("_JSONAssetPool")[__materialName];
            }

            for (let i = 0; i < __textures.length; ++i)
            {
                const textureURL = __textures[i].url;
                let textureType = __textures[i].type || this.GetTextureType(textureURL);

                const textureTypeKey = GetTextureTypeKey(textureType);
                if (!textureTypeKey)
                {
                    if (textureType)
                    {
                        LogMessage.Error("Texture type " + textureType + " not supported (" + textureURL + ")");
                    }
                    else
                    {
                        LogMessage.Error("Texture type could not be determined from its texture name (" + textureURL + ")");
                    }
                    LogMessage.Error("Programmer: choose one of these texture types:'diffuse', 'specular', 'normal_map', 'internal_specular', 'diffraction_grating', or 'diffraction_grating_filter'");
                    continue;
                }

                const textureKey = material.Properties.front[textureTypeKey];
                if (!textureKey)
                {
                    LogMessage.Warning("Texture of type " + textureType + " not found for material " + __materialName + ". This viewer will create one for you.");
                    const textureKey = __materialName + "_" + textureTypeKey + "_" + uuid();//Generate a unique name
                    material.Properties.front[textureTypeKey] = textureKey;
                    this.get("_JSONAssetPool")[textureKey] =
                        {
                            AssetType: "Texture",
                            Properties:
                                {
                                    filepath: textureURL,
                                    transparent: __textures[i].transparent || false,
                                    remapped_alpha_channel: __textures[i].remapped_alpha_channel || false,
                                    swrap: __textures[i].swrap || "GL_REPEAT",
                                    twrap: __textures[i].twrap || "GL_REPEAT"
                                }
                        };
                }
                else
                {
                    const textureAsset = this.get("_JSONAssetPool")[textureKey];
                    textureAsset.Properties.filepath = textureURL;
                    textureAsset.Properties.transparent = __textures[i].transparent || textureAsset.Properties.transparent || false;
                    textureAsset.Properties.remapped_alpha_channel = __textures[i].remapped_alpha_channel || textureAsset.Properties.remapped_alpha_channel || false;
                    textureAsset.Properties.swrap = __textures[i].swrap || textureAsset.Properties.swrap || "GL_REPEAT";
                    textureAsset.Properties.twrap = __textures[i].twrap || textureAsset.Properties.twrap || "GL_REPEAT";
                }

                if (__textures[i].original_width)
                {
                    this.get("_JSONAssetPool")[textureKey].Properties.original_width = __textures[i].original_width;
                }
                if (__textures[i].original_height)
                {
                    this.get("_JSONAssetPool")[textureKey].Properties.original_height = __textures[i].original_height;
                }
            }
        }

    });

    //We define GetMaterials function afterwards since MaterialList is not yet defined and we cannot not call "new MaterialList" otherwise

    /**
     * This convenience function allows the user to retrieve only the listed materials that he/she does want to keep
     * from this materiallist. For example, someone can get the full material list using WebViewerAPI.js::GetMaterialList
     * to retrieve this material list, but then updates one material using SetTextureURL("MyMaterial"); and then he/she
     * wants to call WebViewerAPI::UpdateMaterials using that list.
     * This function takes care of creating a new list with only the sub set of material that are given in __extractionList
     * so that WebViewerAPI::UpdateMaterials can take less time to update all the materials.
     * @param {Array} __extractionList an array of strings, each referring a material in this material list.
     * @returns {MaterialList} a new material list, composed with only the subset of specified materials
     */
    MaterialList.prototype.GetMaterials = function (__extractionList)
    {
        const newMaterialPool = {};
        for (let i = 0; i < __extractionList.length; ++i)
        {
            const materialName = __extractionList[i];
            const material = this.get("_JSONAssetPool")[materialName];
            if (!material)
            {
                LogMessage.Error("Material " + materialName + " not found in list");
                continue;
            }
            if (material.AssetType !== "Material")
            {
                LogMessage.Error("Asset " + materialName + " is not a material");
                continue;
            }

            newMaterialPool[materialName] = $.extend(true, {}, material);//deep copy of material asset
            const types = this.GetTextureTypes(materialName);
            for (let j = 0; j < types.length; ++j)
            {
                const textureKey = material.Properties.front[GetTextureTypeKey(types[j])];
                const texture = this.get("_JSONAssetPool")[textureKey];

                if (!texture)
                {
                    LogMessage.Error("Texture " + textureKey + " not found in list");
                    continue;
                }
                if (texture.AssetType !== "Texture")
                {
                    LogMessage.Error("Asset " + textureKey + " is not a texture");
                    continue;
                }

                newMaterialPool[textureKey] = $.extend(true, {}, texture);//deep copy of texture asset
            }
        }
        return new MaterialList({_JSONAssetPool:newMaterialPool});
    };

    return MaterialList;

});
