import ViewportLayerModel from "@esko/web-viewer/viewport/model/ViewportLayerModel";
import TextSelectionPopoverModel from "./TextSelectionPopoverModel";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import LineworkWrapper from "@esko/web-viewer/utils/LineworkWrapper";
import _ from "underscore";
/**
 * @exports TextSelectionLayerModel
 * @memberof SvgModule
 * @augments ViewportLayerModel
 * @class This class gives the linework to render selected text and show popover.
 * @constructor
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} ID - the ID of the layer model
 * @param {TextSelectionModel} textSelectionModel - Model of the text selection tool
 */
var TextSelectionLayerModel = function(f, ID, textSelectionModel) {
	ViewportLayerModel.call(this, f, ID);
	ListenerUtil.mixin(this, f);

	// Variables
	this._textSelectionModel = textSelectionModel;
	this._textSelectionModel.on("changed", this._emit.bind(this, "modelChanged"));
	this._textSelectionPopover = null;
	this._textSelectionPopoverListener = null;
	this._worldTextFrames = null;
};

// inherit
TextSelectionLayerModel.prototype = Object.create(ViewportLayerModel.prototype);
TextSelectionLayerModel.prototype.constructor = TextSelectionLayerModel;

/**
 * Start showing the textSelection popover
 */
TextSelectionLayerModel.prototype.showPopover = function() {
	if (!this._textSelectionPopover) {
		this._textSelectionPopover = new TextSelectionPopoverModel(this.f, "TextSelectionPopover");
		this._textSelectionPopoverListener = this._registerListener(
			"popupClosed",
			function(e) {
				this.hidePopover();
			},
			this,
			this._textSelectionPopover
		);
		this.modelChanged();

		this.addPopoverModel(this._textSelectionPopover);
	}
};

TextSelectionLayerModel.prototype.hasLinework = function() {
	return true;
}

TextSelectionLayerModel.prototype.getLinework = function() {
	return _.values(_.groupBy(this._textSelectionModel.getSelection(), "lineIndex")).map(line => {
		return line.reduce((mem, el) => {
			var {minX, minY, maxX, maxY} = el;
			var rect = new ShapeUtils.Rect(new ShapeUtils.Point(minX, minY), new ShapeUtils.Size(maxX - minX, maxY - minY));
			if(mem) {
			return mem.calculateArea(rect);
			} else {
				return rect;
			}
		}, null);
	}).map(rect => {
		var lww = {};
		lww.rect = rect;
		lww.fillcolor=  "rgba(51, 153, 255, 0.7)"
		lww.type = -1;
		return lww;
	});

}


/**
 * Hide the textSelection popover
 */
TextSelectionLayerModel.prototype.hidePopover = function() {
	if (this._textSelectionPopover) {
		this.removePopoverModel(this._textSelectionPopover.getID());
		this.removePopover(this._textSelectionPopover.getID());
		this._unregisterListener(this._textSelectionPopoverListener);
		this._textSelectionPopover = null;
	}
};

/**
 * Override the model changed, so the popovermodel is updated when needed.
 * @override
 */
TextSelectionLayerModel.prototype.modelChanged = function() {
	if (this._textSelectionPopover) {
		this._textSelectionPopover.setTextContent(this._textSelectionModel.getContent());
		var worldTextFrames = this.getLinework().map(lw => lw.rect); 
		this._textSelectionPopover.setWorldTextFrames(worldTextFrames);
		this._textSelectionPopover.setPopoverLocation(this._textSelectionModel.getEndPoint());
	}
};



/**
 * Get Popover Model
 * @returns Popover Model
 */
TextSelectionLayerModel.prototype.getPopoverModel = function() {
	return this._textSelectionPopover;
};

export default TextSelectionLayerModel;
