import AbstractModel from "@esko/web-viewer/core/AbstractModel";
import ShapeUtils from "@esko/shapeutils";
import rbush from "rbush";
import rbushknn from "rbush-knn";
import _ from "underscore";

/**
 * @exports TextSelectionModel
 * @memberof SvgModule
 * @class The TextSelectionModel is used for representing all text selection tool info
 * @constructor
 * @augments {Core.AbstractModel}
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {String} privateChannelName - The privateChannelName of the model
 */
var TextSelectionModel = function(f, privateChannelName) {
	AbstractModel.call(this, f, privateChannelName);

	this._startPoint = null;
	this._endPoint = null;
	this._rbush;
};

TextSelectionModel.prototype = Object.create(AbstractModel.prototype);
TextSelectionModel.prototype.constructor = TextSelectionModel;

/**
 * Set the start point.
 * @param {ShapeUtils.Point} startPoint
 */
TextSelectionModel.prototype.setStartPoint = function(startPoint) {
	if(this._startPoint !== startPoint) {
	this._startPoint = startPoint;
	this._emit("changed");
	}

};

/**
 * 
 * @param {Array.<Object>} chars - Array of selectedable characters. A character is an object: 
 * {
 * 	minX: Number,
 *  maxX: Number,
 *  minY: Number,
 *  maxY: Number,
 *  elementIndex: Number, // the read order of the character.
 *  lineIndex: Number // all characters on a line should have the same line index
 *  character: String
 * } 
 */
TextSelectionModel.prototype.setCharacters = function(chars) {
	this._rbush = new rbush();
	this._rbush.load(chars);
	this._sorted = chars.sort((x,y) => {x.elementIndex - y.elementIndex}); // make sure we have a sorted list

}



/**
 * Get the start point.
 * @returns {ShapeUtils.Point} - startPoint
 */
TextSelectionModel.prototype.getStartPoint = function() {
	if (this._startPoint) {
		return this._startPoint;
	} else {
		return new ShapeUtils.Point(0, 0);
	}
};

/**
 * Set the end point.
 * @param {ShapeUtils.Point} endPoint
 */
TextSelectionModel.prototype.setEndPoint = function(endPoint) {
	if(this._endPoint !== endPoint) {
		this._endPoint = endPoint;
		this._emit("changed");
	}
};

/**
 * Get the end point.
 * @returns {ShapeUtils.Point} - endPoint
 */
TextSelectionModel.prototype.getEndPoint = function() {
	if (this._endPoint) {
		return this._endPoint;
	} else {
		return new ShapeUtils.Point(0, 0);
	}
};

/**
 * Returns an Array of characters (as in setCharacters) in order. 
 */
TextSelectionModel.prototype.getSelection = function(){
	if(!this._rbush || !this._sorted || !this._startPoint || !this._endPoint) {
		return [];
	}
	var [start] = rbushknn(this._rbush, this._startPoint.x, this._startPoint.y, 1);
	var [end] = rbushknn(this._rbush, this._endPoint.x, this._endPoint.y, 1);
	if(!start || !end) {
		return [];
	}
	var min = Math.min(start.elementIndex, end.elementIndex);
	var max = Math.max(start.elementIndex, end.elementIndex);
	return _.range(min, max + 1).map(i => this._sorted[i]);
}

/**
 * Get the text content of the selection
 * @returns {String} - content
 */
TextSelectionModel.prototype.getContent = function() {
	return _.values(_.groupBy(this.getSelection(), "lineIndex")).map(line => {
		return line.map(c => c.character).join("");
	}).join(" ");
};

export default TextSelectionModel;
