import _ from "underscore";
import $ from "jquery";
import AbstractTool from "@esko/web-viewer/viewport/tools/AbstractTool";
import CompareConstants from "@esko/web-viewer/compare/CompareConstants";
import CoreConstants from "@esko/web-viewer/core/CoreConstants";
import ConversionConstants from "@esko/web-viewer/utils/ConversionConstants";
import TextSelectionLayerModel from "./TextSelectionLayerModel";
import rbush from "rbush";
import ShapeUtils from "@esko/shapeutils";
import ListenerUtil from "@esko/web-viewer/utils/ListenerUtil";
import rbushknn from "rbush-knn";
var DEFAULTSTATE = "default";
var DRAGGINGSTATE = "dragging";
var SELECTEDSTATE = "selected";

/**
 * @exports TextSelectionTool
 * @memberof SvgModule
 * @class The TextSelectionTool allows to select text in SVG files
 * @constructor
 * @augments {Core.AbstractTool}
 *
 * @param {Core.ViewerFacade} f - Viewer facade
 * @param {SvgModule.TextSelectionModel}
 */
var TextSelectionTool = function(f, textSelectionModel) {
	AbstractTool.call(this, f, CoreConstants.TOOLS.TEXTSELECTION);
	ListenerUtil.mixin(this, f);

	// listen tot popover interaction to change state
	this.currentPopoverListener = null;

	/**
	 * @member {TextSelectionModel}
	 * @private
	 */
	this._textSelectionModel = textSelectionModel;

	/**
	 * @member {TextSelectionLayerModel}
	 * @private
	 */
	this._textSelectionLayerModel = new TextSelectionLayerModel(f, this.getIID(), this._textSelectionModel);

	this.f.whenModuleAvailable(
		"viewport",
		function(viewportModule) {
			this._viewport = viewportModule;
			if (this.isActive()) {
				this._viewport
					.getFirstViewOccurrence("Main")
					.getViewContext()
					.addClass("enableTextSelection");
			}
		},
		this
	);

	/**
	 * @member {Rbush}
	 * @private
	 */
	this._rbush = null;

	f.getDocumentController().whenActiveOrDifferencePageChangedAndParametersAvailableOrChange(_.bind(this._onPageChange, this));


	/**
	 * DEFAULT STATE -> nothing set
	 */
	this.registerState(DEFAULTSTATE, {
		enterState: _.bind(function() {
			this._textSelectionModel.setStartPoint(null);
			this._textSelectionModel.setEndPoint(null);
		}, this),

		leaveState: _.bind(function() {}, this),

		mousedown: _.bind(function(e) {
			document.activeElement && document.activeElement.blur();
			if (!this.isActive()) {
				// This is only triggered in the SELECTED state (destroys popup and text selection)
				this.changeState(DEFAULTSTATE);
			} else {
				this._textSelectionModel.setEndPoint(null);
				var curLoc = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
				this._textSelectionModel.setStartPoint(curLoc);
				this.changeState(DRAGGINGSTATE);
	
			}
		}, this),

		mouseup: function() {},

		mousemove: function() {},

		mouseout: function() {},

		mouseleave: function() {},

		/**
		 * no context menu!
		 */
		contextmenu: _.bind(function(e) {
			e.preventDefault();
			e.stopPropagation();
		}, this),

		canBeInterrupted: true
	});

	/**
	 * Dragging STATE -> selecting text
	 */
	this.registerState(
		DRAGGINGSTATE,
		_.defaults(
			{

				enterState: function(){},
				mousedown: function() {},

				mouseup: _.bind(function(e) {
					var curLoc = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
					this._textSelectionModel.setEndPoint(curLoc);

					if (curLoc.equals(this._textSelectionModel.getStartPoint())) {
						this.changeState(DEFAULTSTATE);
					} else {
						this.changeState(SELECTEDSTATE);
					}
				}, this),

				mousemove: _.bind(function(e) {
					var curLoc = this.keepPointWithinWorldBounds(this.getWorldPoint(e), false);
					// update selected text with current location

					if (!curLoc.equals(this._textSelectionModel.getStartPoint())) {
							textSelectionModel.setEndPoint(curLoc);
					}
				}, this)
			},
			this.getState(DEFAULTSTATE)
		)
	);

	/**
	 * selected STATE -> text is selected and annotation balloon is shown
	 */
	this.registerState(
		SELECTEDSTATE,
		_.defaults(
			{
				enterState: _.bind(function() {

					this._textSelectionLayerModel.showPopover();
					}.bind(this)),

				leaveState: _.bind(function() {
					this._textSelectionLayerModel.hidePopover();
					if (this.currentPopoverListener) {
						this._unregisterListener(this.currentPopoverListener);
					}
				}, this)
			},
			this.getState(DEFAULTSTATE)
		)
	);

	this.registerActiveEvents(["mousedown", "mousemove", "mouseup", "mouseleave", "contextmenu"]);
	this.registerPassiveEvents(["mousedown"]);

	// Esc allows to reset selection
	this.activeKeyEvents.push([
		"esc",
		_.bind(function(e) {
			this.changeState(DEFAULTSTATE);
			e.preventDefault();
			e.stopPropagation();
		}, this),
		"keyup"
	]);

	/**
	 * @override
	 */
	this._override("initialize", function(widget, toolManager) {
		this._super(widget, toolManager);
		this.changeState(DEFAULTSTATE);
	});

	/**
	 * @override
	 */
	this._override("activate", function(interruptRequest) {
		this._super(interruptRequest);
		if (this._viewport) {
			this._viewport
				.getFirstViewOccurrence("Main")
				.getViewContext()
				.addClass("enableTextSelection");
		}
		this._initializeRbush();
		this.f.getModule("viewport").addLayer(this.getIID(), this._textSelectionLayerModel);
	});

	/**
	 * @override
	 */
	this._override("deactivate", function(interruptRequest) {
		this._super(interruptRequest);
		if (!interruptRequest) {
			this._rbush = null;
			this.changeState(DEFAULTSTATE);
			// The SVGLayer should be first in the dom
			this.f.emit("viewport.viewmodel.Main.layers.swapIfIndexIsHigher", "SVGLayer", "SVGDifferenceLayer");
			this._textSelectionLayerModel.modelChanged(true);
			this.f.getModule("viewport").removeLayer(this.getIID());
			if (this._viewport) {
				this._viewport
					.getFirstViewOccurrence("Main")
					.getViewContext()
					.removeClass("enableTextSelection");
			}
		}
	});

	// activate text selection tool
	this.passiveKeyEvents.push([
		"t",
		_.bind(function(e) {
			if (this.f.getActivePage().hasSvg && this.f.getActivePage().hasSvg()) {
				e.preventDefault();
				f.emit("toolManager.activateTool", this);
			}
		}, this),
		"keypress"
	]);
};

// extend from the AbstractTool
// inherit
TextSelectionTool.prototype = Object.create(AbstractTool.prototype);
TextSelectionTool.prototype.constructor = TextSelectionTool;

TextSelectionTool.prototype._onPageChange = function() {
	if (this.isActive()) {
		this.changeState(DEFAULTSTATE);
		this._rbush = null;
		this._initializeRbush();
	}
};

/**
 * Returns the default cursor for the tool
 * @returns {String}
 */
TextSelectionTool.prototype.getDefaultCursor = function() {
	return "cursor-textSelect";
};

/**
 * If the world is animating, cancel it and wait for the svglayer to have been repainted
 */
TextSelectionTool.prototype._initializeRbush = function() {
	if (this._rbush) {
		return;
	}
	var svgLayer;
	if (this._widget.getWorld().isAnimating()) {
		// As there is a lot of asynchronicity in the viewer, we have to make sure the SVG's transform is the same as the viewer's transform
		// To make sure this is the case we first have to listen to the a repaint of the view.
		// This will trigger all layers to rerender. This also happens asynchronous. So we should then listen to the new repaint of the svg layer
		svgLayer = this._viewport.getFirstViewOccurrence("Main").getLayer("SVGLayer");
		var viewListener = this._registerListener(
			"repainted",
			function() {
				// The widget has been repainted, now we have to wait to make sure the svg's transformation is updated as well
				var layerListener = this._registerListener(
					"repainted",
					function() {
						this._calculateRbush();
						this._unregisterListener(layerListener);
					},
					this,
					svgLayer
				);
				this._unregisterListener(viewListener);
			},
			this,
			this._widget
		);

		this._widget.getWorld().cancelAnimation();
	} else {
		this._calculateRbush();
	}
};

/**
 * Initializes the Rbush with the text elements in the svg
 */
TextSelectionTool.prototype._calculateRbush = function() {
	var rbushElems = [];
	var svgLayer = this._viewport
		.getFirstViewOccurrence("Main")
		.getLayer("SVGLayer")
		.getLayerElement();
	if (this.f.getDocumentController().hasDifferenceDocumentPage()) {
		var svgDifferenceLayer = this._viewport
			.getFirstViewOccurrence("Main")
			.getLayer("SVGDifferenceLayer")
			.getLayerElement();
		if (!this.f.getActivePage().get("isRightPage")) {
			// For the text selection to work in readerspread mode the left page should be the first in the dom.
			this.f.emit("viewport.viewmodel.Main.layers.swapIfIndexIsHigher", "SVGLayer", "SVGDifferenceLayer");

			this._calculateRbushElements($(svgLayer), rbushElems);
			this._calculateRbushElements($(svgDifferenceLayer), rbushElems);
		} else {
			// For the text selection to work in readerspread mode the left page should be the first in the dom.
			this.f.emit("viewport.viewmodel.Main.layers.swapIfIndexIsHigher", "SVGDifferenceLayer", "SVGLayer");
			this._calculateRbushElements($(svgDifferenceLayer), rbushElems);
			this._calculateRbushElements($(svgLayer), rbushElems);
		}
	} else {
		this._calculateRbushElements($(svgLayer), rbushElems);
	}

	try {
		this._textSelectionModel.setCharacters(rbushElems);
	} catch (e) {
		console.error("rbush - could not insert character bounding boxes");
	}

	if (process.env.NODE_ENV === "development") {
		//this._renderRbush(); //debug
	}
};

/**
 * @param {Node} svgElement
 * @param {Array.<Object>} rbushElems
 */
TextSelectionTool.prototype._calculateRbushElements = function(svgElement, rbushElems) {
	var self = this;
	var lineIndex = 0;
	svgElement.find("text").each(function(index) {
		var el = $(this);
		var length = el[0].getNumberOfChars();
		var textContentTrimmed = this.textContent.trim();

		var getChar = (i) => {
			var coords = self._convertCoords(i, el[0]);
			var obj = {
				minX: coords.origin.x,
				minY: coords.origin.y,
				maxX: coords.origin.x + coords.size.width,
				maxY: coords.origin.y + coords.size.height,
				elementIndex: rbushElems.length,
				character: textContentTrimmed[i],
				lineIndex: lineIndex
			};
			return obj
		}

		for (var j = 0; j < length; j++) {
			let obj = getChar(j);
			if(obj.character) {
				rbushElems.push(obj);
			}
		}
		lineIndex++;
		
	});
};

/**
 * @param {Integer} index of character in text element
 * @param {Node} element - the text element
 */
TextSelectionTool.prototype._convertCoords = function(index, element, svg) {
	var extent = element.getExtentOfChar(index);
	var matrix = element.getScreenCTM();
	svg = this._viewport
		.getFirstViewOccurrence("Main")
		.getLayer("SVGLayer")
		.getLayerElement();

	var pt = svg[0].createSVGPoint();
	pt.x = extent.x;
	pt.y = extent.y;
	pt = pt.matrixTransform(matrix);

	var pt2 = svg[0].createSVGPoint();
	pt2.x = extent.x + extent.width;
	pt2.y = extent.y + extent.height;
	pt2 = pt2.matrixTransform(matrix);

	var parent = this._viewport
		.getFirstViewOccurrence("Main")
		.getViewContext()
		.offset();
	var scenePoint = new ShapeUtils.Point(pt.x - parent.left, pt.y - parent.top);
	var sceneSize = new ShapeUtils.Size(pt2.x - pt.x, pt2.y - pt.y);

	return this.sceneToWorldRect(new ShapeUtils.Rect(scenePoint, sceneSize));
};

TextSelectionTool.prototype.canBeInterrupted = function(tool) {
	return this.getCurrentStateName() !== DRAGGINGSTATE;
};

if (process.env.NODE_ENV === "development") {
	/**
	 * Renders rectangles around the text elements
	 * @private
	 */
	TextSelectionTool.prototype._renderRbush = function() {
		this.requestCanvasLayer("Main");
		var ctx = this.getCanvasCtx();
		if (!this._interval) {
			this._interval = setInterval(
				_.bind(function() {
					if (this._rbush) {
						var allElements = this._rbush.all();
						ctx.setTransform(1, 0, 0, 1, 0, 0);
						ctx.clearRect(0, 0, ctx.canvas.width, ctx.canvas.height);
						ctx.beginPath();
						_.each(
							allElements,
							function(obj) {
								var worldOrigin = new ShapeUtils.Point(obj.minX, obj.minY);
								var worldSize = new ShapeUtils.Size(obj.maxX - obj.minX, obj.maxY - obj.minY);
								var sceneOrigin = this.worldToScenePoint(worldOrigin);
								var sceneSize = this.worldToSceneSize(worldSize);
								ctx.rect(sceneOrigin.x, sceneOrigin.y, sceneSize.width, sceneSize.height);
							},
							this
						);

						ctx.stroke();
					}
				}, this),
				1000 / 30
			);
		}
	};
}

/**
 * Determines if this tool can be activated
 *
 * @returns  {Boolean}
 */
TextSelectionTool.prototype.canBeActivated = function() {
	var compareModule = this.f.getModule("compare");
	if (compareModule) {
		var mode = compareModule.getActiveMode();
		var allowedModes = [CompareConstants.mode.SHOW_CURRENT, CompareConstants.mode.READERSPREAD];
		return !mode || allowedModes.indexOf(mode.getName()) !== -1;
	}
	return true;
};

export default TextSelectionTool;
